import { Request, Response } from "express";
import { User } from "../models/user";
import {
  saveSession,
  hashPassword,
  decodeBase64,
  matchPassword,
  getUserSession,
} from "../utils/utils";
import { isUsernameValid, isPasswordValid } from "../validation/validation";
import { Session } from "../models/session";
import { MONGODB_IGNORED_FIELDS } from "../utils/constants";

export async function SignUp(req: Request, res: Response) {
  try {
    const { accountName, username, password } = req.body;

    if (!accountName || !username || !password) {
      return res.status(400).json({ status: "err" });
    }

    const existingUser = await User.findOne({ accountName })
      .select(MONGODB_IGNORED_FIELDS)
      .lean();

    if (existingUser) {
      return res.status(400).json({ status: 1 });
    }

    if (!isPasswordValid(password)) {
      return res.status(400).json({ status: "err" });
    }

    const decodedPassword = decodeBase64(password);
    const hashedPassword = await hashPassword(decodedPassword);

    const user = new User({
      accountName: accountName,
      username: username,
      password: hashedPassword,
    });

    user
      .save()
      .then(() => saveSession(res, accountName, username))
      .catch((err) => {
        console.log(err);
        res.status(500).json({ status: "err" });
      });
  } catch (error) {
    console.error("error on signup:", error);
    res.status(500).json({ status: "err" });
  }
}

export async function Login(req: Request, res: Response) {
  try {
    const { AccountName, Password } = req.body;

    if (!AccountName || !Password) {
      return res.status(400).json({ status: "err" });
    }

    const user = await User.findOne({ accountName: AccountName })
      .select(`accountName username password -_id`)
      .lean();

    if (!user) {
      return res.status(401).json({ status: "err" });
    }

    if (
      user.accountName === null ||
      user.accountName === undefined ||
      user.password === undefined ||
      user.password === null
    ) {
      return res.status(401).json({ status: "err" });
    }

    const decodedPassword = decodeBase64(Password);
    const isPasswordValid = await matchPassword(decodedPassword, user.password);

    if (!isPasswordValid) {
      return res.status(401).json({ status: "err" });
    }

    saveSession(res, user.accountName, user.username);
  } catch (error) {
    console.error("error on login:", error);
    res.status(500).json({ status: "err" });
  }
}

export async function ChangeUsername(req: Request, res: Response) {
  try {
    const { newUsername } = req.body;

    if (!newUsername || !isUsernameValid(newUsername)) {
      return res.status(400).json({ status: "err" });
    }

    const session = await getUserSession(req, "accountName");

    if (!session) {
      return res.status(401).json({ status: "err" });
    }

    await User.updateOne(
      { accountName: session.accountName },
      { $set: { username: newUsername } }
    );

    res.status(200).json({ status: "ok" });
  } catch (error) {
    console.error("error on changing username", error);
    res.status(500).json({ status: "err" });
  }
}

export async function ChangePassword(req: Request, res: Response) {
  try {
    // cp = current password
    // np = new password
    const { cp, np } = req.body;

    if (!cp || !np) {
      return res.status(400).json({ status: "err" });
    }

    console.log(cp, np);

    const session = await getUserSession(req, "accountName");

    if (!session) {
      return res.status(401).json({ status: "err" });
    }

    const accountName = session.accountName;

    const decodedCurrentPassword = decodeBase64(cp);

    const existingUser = await User.findOne({
      accountName: accountName,
    })
      .select(`password -_id`)
      .lean();

    if (!existingUser) {
      return res.status(400).json({ status: 1 });
    }

    const decodedNewPassword = decodeBase64(np);

    if (!isPasswordValid(decodedCurrentPassword)) {
      return res.status(400).json({ status: "err" });
    }

    const isPasswordMatching = await matchPassword(
      decodedCurrentPassword,
      existingUser.password
    );

    if (!isPasswordMatching || !isPasswordValid(decodedNewPassword)) {
      return res.status(401).json({ status: "err" });
    }

    await User.updateOne(
      { accountName: accountName },
      { $set: { password: await hashPassword(decodedNewPassword) } }
    );

    res.status(200).json({ status: "ok" });

    // delete all sessions
    await Session.deleteMany({ accountName: accountName });
  } catch (error) {
    console.error("error on changing password", error);
    res.status(500).json({ status: "err" });
  }
}

export async function Logout(req: Request, res: Response) {
  try {
    const session = await getUserSession(req);

    if (!session) {
      return res.status(401).json({ status: "err" });
    }

    await Session.deleteOne({ sessionId: session.sessionId });

    res.status(200).json({ status: "ok" });
  } catch (error) {
    console.error("error on logout", error);
    res.status(500).json({ status: "err" });
  }
}
